#ifndef COMPOSE_HH
#define COMPOSE_HH
#include "paramlists.hh"
#include "argsets.hh"
#include "void_return.hh"
#include "exprtype.hh"
#include "argset_transform.hh"
#include "argset_ncref.hh"
#include "funcopimpl.hh"
#include "rthelper.hh"

namespace FuncImpl
{

    template<typename Func,typename ArgSet>
    class ComposedFunc;

    template<typename Func,typename ArgSet,typename CAS>
    class ComposedHelper;
}

namespace ArgSets
{
    // specialize/overload ArgSet functions for ComposedFunc

    template<typename Func,typename ArgSet,typename CAS>
    struct TransformArgSetResult<FuncImpl::ComposedFunc<Func,ArgSet>,CAS>
    {
	typedef typename FuncImpl::ComposedHelper<Func,ArgSet,CAS>::Type Type;
    };

    template<typename Func,typename ArgSet,typename CAS>
    typename FuncImpl::ComposedHelper<Func,ArgSet,CAS>::Type transformArg(const FuncImpl::ComposedFunc<Func,ArgSet>& cf,const CAS& cas)
    {
	return cf.call(cas);
    }

    template<typename Func,typename ArgSet>
    struct NumArgs<FuncImpl::ComposedFunc<Func,ArgSet> >
    {
	static const unsigned result=MinArgs<ArgSet>::result;
    };

    template<unsigned i,typename CAS>
    struct TransformArgSetResult<Holder::RefHolder<FuncImpl::ParamType<i> >,CAS>
    {
	typedef typename GetASEntry<CAS,i>::Type& Type;
    };
    
    template<unsigned i,typename CAS>
    typename GetASEntry<CAS,i>::Type& transformArg(const Holder::RefHolder<FuncImpl::ParamType<i> >& arg,const CAS& cas)
    {
	return GetASEntry<CAS,i>::getVal(cas);
    }

    template<unsigned i>
    struct NumArgs<Holder::RefHolder<FuncImpl::ParamType<i> > >
    {
	static const unsigned result=i;
    };
}


namespace FuncImpl
{
    // specialize RTHelper for ComposedFunc
    template<typename Func,typename ArgSet1,
	     typename ArgSet2>
    struct RTHelper<ComposedFunc<Func,ArgSet1>,ArgSet2>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet1,ArgSet2>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
    };

    template<typename Func,typename ArgSet,typename CAS,unsigned numArgs=ArgSets::CountArgs<ArgSet>::result>
    struct CallerHelper
    {};

    // specialize for a function call with each number of arguments
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,0>
    {
	typedef typename RTHelper<Func,ArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet&,const CAS&)
	{
	    return func();
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,1>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;

	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,2>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,3>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,3>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,4>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,3>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,4>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,5>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,3>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,4>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,5>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,6>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,3>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,4>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,5>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,6>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,7>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,3>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,4>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,5>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,6>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,7>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,8>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,3>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,4>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,5>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,6>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,7>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,8>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,9>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,3>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,4>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,5>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,6>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,7>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,8>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,9>::getVal(as),cas));
	}
    };
    template<typename Func,typename ArgSet,typename CAS>
    struct CallerHelper<Func,ArgSet,CAS,10>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
	static Type callFunc(const Func& func,const ArgSet& as,const CAS& cas)
	{
	    return func(transformArg(ArgSets::GetASEntry<ArgSet,1>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,2>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,3>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,4>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,5>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,6>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,7>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,8>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,9>::getVal(as),cas),
			transformArg(ArgSets::GetASEntry<ArgSet,10>::getVal(as),cas));
	}
    };
    
    template<typename Func,typename ArgSet,typename CAS>
    struct ComposedHelper
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<Func,TransformedArgSet>::Type Type;
    };

    template<typename Func,typename ArgSet>
    struct ComposedHelper<Func,ArgSet,ArgSets::ArgSet0>
    {
	typedef typename CT::If<(ArgSets::MinArgs<ArgSet>::result==0),
                                ArgSets::TransformArgSetHelper<ArgSet,ArgSets::ArgSet0>,
                                CT::Identity<void> >::Type::Type TransformedArgSet;
	typedef typename CT::If<(ArgSets::MinArgs<ArgSet>::result==0),
                                RTHelper<Func,TransformedArgSet>,
                                CT::Identity<void> >::Type::Type Type;

    };

    template<typename Func,typename ArgSet>
    struct CFHelper
    {
	template<typename CAS>
	struct Helper
	{
	    typedef typename FuncImpl::ComposedHelper<Func,ArgSet,CAS>::Type ReturnType;
	    typedef typename FuncImpl::ComposedHelper<Func,ArgSet,CAS>::TransformedArgSet TransformedArgSet;
	};    

	typedef ComposedFunc<Func,ArgSet> const& DerivedType;
    };
    

    template<typename Func,typename ArgSet>
    struct ComposedFunc:
	public FunctionCallOperatorImpl<CFHelper<Func,ArgSet>,typename ArgSets::FindNonConstArgs<ArgSet>::Type>
    {
	template<typename CAS>
	typename CFHelper<Func,ArgSet>::template Helper<CAS>::ReturnType call(const CAS& cas) const
	{
	    typedef CallerHelper<Func,ArgSet,CAS> CallerType;
	    return CallerType::callFunc(func,as,cas);
	}

	Func func;
	ArgSet as;

	ComposedFunc(const Func& func_,const ArgSet& as_):
	    func(func_),as(as_)
	{}
    };


}

namespace Compose
{

template<typename Func>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet0 > compose(Func func)
{
    typedef ArgSets::ArgSet0 ArgSet;
    ArgSet as;
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet1<Arg1> > compose(Func func,const Arg1& a1)
{
    typedef ArgSets::ArgSet1<Arg1> ArgSet;
    ArgSet as(a1);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet2<Arg1,Arg2> >
compose(Func func,const Arg1& a1,const Arg2& a2)
{
    typedef ArgSets::ArgSet2<Arg1,Arg2> ArgSet;
    ArgSet as(a1,a2);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2,typename Arg3>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet3<Arg1,Arg2,Arg3> >
compose(Func func,const Arg1& a1,const Arg2& a2,const Arg3& a3)
{
    typedef ArgSets::ArgSet3<Arg1,Arg2,Arg3> ArgSet;
    ArgSet as(a1,a2,a3);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2,typename Arg3,typename Arg4>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet4<Arg1,Arg2,Arg3,Arg4> >
compose(Func func,const Arg1& a1,const Arg2& a2,const Arg3& a3,const Arg4& a4)
{
    typedef ArgSets::ArgSet4<Arg1,Arg2,Arg3,Arg4> ArgSet;
    ArgSet as(a1,a2,a3,a4);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2,typename Arg3,typename Arg4,typename Arg5>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet5<Arg1,Arg2,Arg3,Arg4,Arg5> >
compose(Func func,const Arg1& a1,const Arg2& a2,const Arg3& a3,const Arg4& a4,const Arg5& a5)
{
    typedef ArgSets::ArgSet5<Arg1,Arg2,Arg3,Arg4,Arg5> ArgSet;
    ArgSet as(a1,a2,a3,a4,a5);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2,typename Arg3,typename Arg4,typename Arg5,
	 typename Arg6>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet6<Arg1,Arg2,Arg3,Arg4,Arg5,
			  Arg6> >
compose(Func func,const Arg1& a1,const Arg2& a2,const Arg3& a3,const Arg4& a4,const Arg5& a5,
	const Arg6& a6)
{
    typedef ArgSets::ArgSet6<Arg1,Arg2,Arg3,Arg4,Arg5,
	Arg6> ArgSet;
    ArgSet as(a1,a2,a3,a4,a5,
	      a6);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2,typename Arg3,typename Arg4,typename Arg5,
	 typename Arg6,typename Arg7>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet7<Arg1,Arg2,Arg3,Arg4,Arg5,
			  Arg6,Arg7> >
compose(Func func,const Arg1& a1,const Arg2& a2,const Arg3& a3,const Arg4& a4,const Arg5& a5,
	const Arg6& a6,const Arg7& a7)
{
    typedef ArgSets::ArgSet7<Arg1,Arg2,Arg3,Arg4,Arg5,
	Arg6,Arg7> ArgSet;
    ArgSet as(a1,a2,a3,a4,a5,
	      a6,a7);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2,typename Arg3,typename Arg4,typename Arg5,
	 typename Arg6,typename Arg7,typename Arg8>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet8<Arg1,Arg2,Arg3,Arg4,Arg5,
			  Arg6,Arg7,Arg8> >
compose(Func func,const Arg1& a1,const Arg2& a2,const Arg3& a3,const Arg4& a4,const Arg5& a5,
	const Arg6& a6,const Arg7& a7,const Arg8& a8)
{
    typedef ArgSets::ArgSet8<Arg1,Arg2,Arg3,Arg4,Arg5,
	Arg6,Arg7,Arg8> ArgSet;
    ArgSet as(a1,a2,a3,a4,a5,
	      a6,a7,a8);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2,typename Arg3,typename Arg4,typename Arg5,
	 typename Arg6,typename Arg7,typename Arg8,typename Arg9>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet9<Arg1,Arg2,Arg3,Arg4,Arg5,
			  Arg6,Arg7,Arg8,Arg9> >
compose(Func func,const Arg1& a1,const Arg2& a2,const Arg3& a3,const Arg4& a4,const Arg5& a5,
	const Arg6& a6,const Arg7& a7,const Arg8& a8,const Arg9& a9)
{
    typedef ArgSets::ArgSet9<Arg1,Arg2,Arg3,Arg4,Arg5,
	Arg6,Arg7,Arg8,Arg9> ArgSet;
    ArgSet as(a1,a2,a3,a4,a5,
	      a6,a7,a8,a9);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}

template<typename Func,typename Arg1,typename Arg2,typename Arg3,typename Arg4,typename Arg5,
	 typename Arg6,typename Arg7,typename Arg8,typename Arg9,typename Arg10>
FuncImpl::ComposedFunc<Func,ArgSets::ArgSet10<Arg1,Arg2,Arg3,Arg4,Arg5,
			   Arg6,Arg7,Arg8,Arg9,Arg10> >
compose(Func func,const Arg1& a1,const Arg2& a2,const Arg3& a3,const Arg4& a4,const Arg5& a5,
	const Arg6& a6,const Arg7& a7,const Arg8& a8,const Arg9& a9,const Arg10& a10)
{
    typedef ArgSets::ArgSet10<Arg1,Arg2,Arg3,Arg4,Arg5,
	Arg6,Arg7,Arg8,Arg9,Arg10> ArgSet;
    ArgSet as(a1,a2,a3,a4,a5,
	      a6,a7,a8,a9,a10);
    
    return FuncImpl::ComposedFunc<Func,ArgSet>(func,as);
}


}

#endif
